from numpy import sqrt,exp,array,argmax

def cubic(a2,a1,a0):
  #gives the real solution to the cubic equation
  #(Wolfram)
  Q = (3*a1-a2**2)/9
  R = (9*a2*a1 -27*a0 -2*a2**3)/54
  D = Q**3 +R**2

  #now the number become complex
  S = (R+sqrt(D+0j))**(1./3)
  T = (R-sqrt(D+0j))**(1./3)

  #only the first solution is real; return this one
  x1 = -a2/3 +(S+T)
  #x2 = -a2/3 -(S+T)/2 +0.5j*sqrt(3)*(S-T)
  #x3 = -a2/3 -(S+T)/2 -0.5j*sqrt(3)*(S-T)
  return x1


def v_xy(x,St,zeta_w):
  #returns the velocities
  vx = -2*zeta_w *St/(1+St**2)
  vy = -zeta_w/(1+St**2) -1.5*x
  return [vx, vy]

def impactB(St, zeta_w, al):
  bset=0
  bhyp=0
  b3b=0
 #St_crit = min(12/zeta_w**3, 1)
  St_crit = min(12/zeta_w**3, 2)
  if St<=St_crit:
    bset = cubic(2*zeta_w/3, 0, -8*St) *exp(-(St/St_crit)**0.65)
  elif St>=St_crit and St>=zeta_w:
    b3b = (2*0.85*al**0.5 +1/St) *exp(-(0.7*zeta_w/St)**5)
  else:#intermediate case (give all)
    va = zeta_w/(1+St**2) *sqrt(1+4*St**2) #approach velocity
    bhyp = al *sqrt(1+6/(al*va**2))
    b3b = (2*0.85*al**0.5 +1/St) *exp(-(0.7*zeta_w/St)**5)
    bset = cubic(2*zeta_w/3, 0, -8*St) *exp(-(St/St_crit)**0.65)

 #This is old
 #
 #if St>=1 and St>=zeta_w:
 #  b3b = 2*0.85*al**0.5 +1/St
 #elif St<St_crit:
 #  bset = cubic(2*zeta_w/3, 0, -8*St) *exp(-(St/St_crit)**0.65)
 #else:
 #  bset = cubic(2*zeta_w/3, 0, -8*St) *exp(-(St/St_crit)**0.65)
 #  va = zeta_w/(1+St**2) *sqrt(1+4*St**2) #approach velocity
 #  bhyp = al *sqrt(1+6/(al*va**2))

  #make sure it returns a real
  bset=bset.real
  return [bset, bhyp, b3b]

def CR_ana(St, zeta_w, al):
  """
  returns [Rcol, va]

  returns the analalytical collision radius and approach velocity 
  in dimensionless units

  inputs:
    St      :: Stokes number
    zeta_w  :: dimensionless headwind
    al      :: dimensionless radius
  """
  #analytical collision rate
  #Approach velocity

  [bset, bhyp, b3b] = impactB(St, zeta_w, al)
  if b3b<>0:#3-body regime
    [vx, vy] = v_xy(2.5, St, zeta_w)
    va = 3.2
    Rcol = b3b
  elif bhyp>bset:#hyperbolic regime
    Rcol = bhyp
    va = zeta_w/(1+St**2) *sqrt(1+4*St**2) #approach velocity
  else:#settling regime
    Rcol = bset
    va = zeta_w/(1+St**2)+3*bset/4 
    [vx, vy] = v_xy(bset, St, zeta_w)
    va = sqrt(vx**2 +vy**2)

  return [Rcol, va]


def P_cols(St, zeta_w, al):
  """
  returns [bsets, vsets]

  returns the analalytical collision radius and approach velocity 
  in dimensionless units

  inputs:
    St      :: Stokes number
    zeta_w  :: dimensionless headwind
    al      :: dimensionless radius
  """
  #analytical collision rate
  #Approach velocity

  [bset, bhyp, b3b] = impactB(St, zeta_w, al)
  v3b = 3.2
  vhyp = zeta_w/(1+St**2) *sqrt(1+4*St**2)
 #[vx, vy] = v_xy(bset, St, zeta_w)
 #vset = sqrt(vx**2 +vy**2)
  vset = 3*bset/2 +zeta_w #approach velocity

  return [[bset,bhyp,b3b],[vset,vhyp,v3b]]


def P_col(St, zeta_w, al):
  """
  Obtain the (dimensionless) collision rate in the 2D regime:
    
    Pcol = 2R_col va
    (collision rate per unit sigma_d)
  """
  [barr, varr] = P_cols(St, zeta_w, al)
  Pcols = 2*array(barr)*array(varr)
  #pick the largest Pcol
  id = argmax(Pcols)

  return [Pcols[id], barr[id]]


def P_colold(St, zeta_w, al):
  """
  Obtain the (dimensionless) collision rate in the 2D regime:
    
    Pcol = 2R_col va
    (collision rate per unit sigma_d)
  """
  [Rcol, va] = CR_ana(St, zeta_w, al)
  return [2*Rcol*va, Rcol]

